/**
  *  \file client/act-ship.cc
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/actions.h"
#include "client/scr-sim.h"
#include "ui/msgbox.h"
#include "util/nls.h"
#include "game/sim-game.h"
#include "game/ship.h"
#include "cpluslib/format.h"
#include "game/cmds.h"
#include "game/racename.h"
#include "util/misc.h"
#include "game/planet.h"
#include "ui/util.h"
#include "client/dlg-cargo.h"
#include "game/global.h"
#include "client/widgets/objlist.h"
#include "game/bum.h"
#include "game/host.h"
#include "game/config.h"
#include "game/cargotr.h"
#include "game/fleet.h"
#include "game/planettransfer.h"
#include "game/shiptransfer.h"

void
doShipAddToSim(GShip& ship, bool ask)
{
    GSimShip* p = WSimScreen::sim_state.getShipById(ship.getId());
    if (p != 0) {
        // Get permission for overwrite
        if (ask && !UIMessageBox(format(_("The simulation already contains a ship with Id #%d. "
                                          "Replace it?"), ship.getId()),
                                 _("Add to Simulation")).doYesNoDialog())
            return;
    } else {
        // Add new ship
        GSimState::ship_index_t i = WSimScreen::sim_state.addShip();
        p = &WSimScreen::sim_state.getShip(i);
    }

    GSimulatorRealGameInterface::copyShip(*p, ship);
    WSimScreen::sim_state.doUpdates();
}

/** Generic "give unit" interface.
    \param cmds  Command container
    \param cmd   Command to generate
    \param id    Id of unit to process
    \param owner Current owner of unit
    \param title Title of dialog
    \return true if user changed something */
static bool
doGive(GCommandContainer& cmds,
       GCommand::Type cmd,
       int id,
       int owner,
       const char* title)
{
    /* Generate list */
    UIStringList list;
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        if (i == owner)
            list.add(i, format(_("%X - (don't give it away)"), i));
        else
            list.add(i, format("%X - %s", i, player_racenames.getShortName(i)));
    }

    /* Generate widget */
    UIStandardListbox listbox(0, NUM_PLAYERS, getTextWidth(*fonts[FONT_NORMAL], list) + 20, list);
    listbox.scrollTo(owner-1, 1);
    if (const GCommand* c = cmds.getCommand(cmd, id)) {
        int val;
        if (stringValue(c->getArg(), val))
            listbox.scrollTo(val-1, 1);
    }

    /* Dialog */
    if (listbox.doStandardListbox(title, string_t())) {
        int new_owner = listbox.getKey();
        if (new_owner == owner) {
            cmds.removeCommand(cmd, id);
        } else {
            cmds.setCommand(cmd, id, format("%d", new_owner));
        }
        return true;
    } else {
        return false;
    }
}

void
doGiveShip(GShip& ship, GCommandContainer& cmds)
{
    if (doGive(cmds, GCommand::phc_GiveShip, ship.getId(), ship.getRealOwner(), _("Give Ship To..."))) {
        ship.markDirty();
    }
}

void
doGivePlanet(GPlanet& planet, GCommandContainer& cmds)
{
    if (doGive(cmds, GCommand::phc_GivePlanet, planet.getId(), planet.getOwner(), _("Give Planet To..."))) {
        planet.markDirty();
    }
}

void
doShipTransferFor(GShip& ship, GObjectReference chosen)
{
    GCargoTransfer tx;
    tx.addNewContainer(new GShipTransfer(ship));

    if (!chosen.isGameObject()) {
        /* jettison */
        tx.addNewContainer(new GShipTransporterTransfer(ship));
    } else if (GShip* s = dynamic_cast<GShip*>(&*chosen)) {
        if (s->getOwner() == ship.getOwner()) {
            /* ship <-> ship */
            ASSERT(s->hasFullShipData());
            tx.addNewContainer(new GShipTransfer(*s));
        } else {
            /* ship <-> foreign ship */
            if (ship.isTransporterInUse(GShip::t_Transfer)) {
                messageBox(_("The cargo transporter is already in use."), _("Cargo Transfer"));
                return;
            }
            if (ship.getTransporter(GShip::t_Transfer).id != 0
                && ship.getTransporter(GShip::t_Transfer).id != s->getId()
                && ship.isTransporterActive(GShip::t_Transfer))
            {
                messageBox(format(_("The cargo transporter is already transferring cargo "
                                    "to ship #%d."),
                                  ship.getTransporter(GShip::t_Transfer).id), _("Cargo Transfer"));
                // FIXME: state name of current target ship here
                // FIXME: we could cancel the transfer here
                return;
            }
            // ship.cancelTranporter(true);
            tx.addNewContainer(new GShipTransporterTransfer(ship, GShip::t_Transfer, *s));
        }
    } else if (GPlanet* p = dynamic_cast<GPlanet*>(&*chosen)) {
        if (p->getOwner() == ship.getOwner() && p->hasFullPlanetData()) {
            tx.addNewContainer(new GPlanetTransfer(*p));
        } else {
            if (ship.isTransporterInUse(GShip::t_Unload)) {
                messageBox(_("The cargo transporter is already in use."), _("Cargo Transfer"));
                return;
            }
            tx.addNewContainer(new GShipTransporterTransfer(ship, GShip::t_Unload, *p));
        }
    } else {
        FAIL();
    }
    doCargoTransfer(tx);
}

static void
doShipBeamUpMulti(GShip& ship, GPlanet& planet)
{
    GCargoTransfer tx;
    tx.addNewContainer(new GShipBumTransfer(getDisplayedTurn(), ship.getId()));
    tx.addNewContainer(new GPlanetBumTransfer(getDisplayedTurn(), planet, ship.getId()));
    doCargoTransfer(tx);
}

void
doShipCargoTransfer(GUniverse& univ, GShip& ship)
{
    GObjectList list;
    int pid = univ.getPlanetAt(ship.getPos());
    if (pid) {
        list.addObject(current_anyplanet_type, pid);
        if (host.isPHost() && config.AllowBeamUpMultiple() && !isMissionLocked(ship, 0)) {
            list.addNonObject(-2, _("Beam up multiple"));
        }
    } else {
        list.addNonObject(-1, _("Jettison into space"));
    }

    for (int i = univ.ty_any_ships.findNextIndex(0); i != 0; i = univ.ty_any_ships.findNextIndex(i)) {
        if (i != ship.getId() && univ.getShip(i).getPos() == ship.getPos())
            list.addObject(current_anyship_type, i);
    }

    GObjectReference chosen = WObjectList(0, 15, 320, list).doStandardDialog2(_("Transfer cargo to..."), "pcc2:cargo");
    if (chosen.isValid()) {
        if (chosen->getId() == -2) {
            doShipBeamUpMulti(ship, univ.getPlanet(pid));
        } else {
            doShipTransferFor(ship, chosen);
        }
    }
}
