/**
  *  \file game/db.cc
  *  \brief Starchart Database I/O
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include "game/db.h"
#include "cpluslib/format.h"
#include "cpluslib/misc.h"
#include "game/drawing.h"
#include "game/global.h"
#include "game/mine.h"
#include "game/planet.h"
#include "game/ship.h"
#include "game/unitscore.h"
#include "int/data.h"
#include "int/intglobal.h"
#include "int/names.h"
#include "io/io.h"
#include "io/streams.h"
#include "util/console.h"
#include "util/except.h"
#include "util/nls.h"

/** Starchart database record numbers. */
enum {
    rPlanetHistory  = 1,
    rShipHistory    = 2,
    rShipTrack      = 3,
    rMinefield      = 4,
    rPainting       = 5,
    rAutoBuild      = 6,
    rShipProperty   = 7,
    rPlanetProperty = 8,
    // PCC 1.1.6+:
    rShipScore      = 9,
    rPlanetScore    = 10,
    rPaintingTags   = 11,
    // PCC 1.1.7+:
    rUfoHistory     = 12
};

static void
loadPropertyRecord(Stream& s,
                   int max_id,
                   IntDataSegment& get(int),
                   const IntVariableNames& db_names,
                   IntVariableNames& live_names)
{
    // Header: Id + count
    char header[4];
    if (s.read(header, 4) != 4) {
        console.write(LOG_WARN, _("Property record has unexpected size"));
        return;
    }
    uint16 id    = getUint16(header);
    uint16 count = getUint16(header+2);
    if (id <= 0 || id > max_id) {
        console.write(LOG_WARN, format(_("Property record has invalid Id (%d) and has been ignored"), id));
        return;
    }

    // Read data into temporary store first
    IntDataSegment tmp;
    tmp.load(s, count);

    // Copy to game object
    IntDataSegment& target = get(id);
    uint32_t limit = std::min(uint32_t(count), uint32_t(db_names.getNumNames()));
    for (uint32 i = 0; i < limit; ++i) {
        IntValue* db_value = tmp[i];
        string_t  db_name  = db_names.lookupAddress(i);

        // Address remapping
        IntVariableNames::index_t live_index = live_names.lookup(db_name);
        if (db_value != 0 && live_index == live_names.nil)
            live_index = live_names.add(db_name);

        // Store value
        if (live_index != live_names.nil)
            target.set(live_index, db_value);
    }
}

static bool
readUnitScoreHeader(Stream& s, TUnitScoreHeader& ush)
{
    /* Read size field */
    TInt16 size;
    if (!getStructure(s, size))
        return false;
    if (size.value < ush.size)
        return false;

    /* Read unit score header */
    if (!getStructure(s, ush))
        return false;

    /* Skip possible trailing data */
    if (size.value > ush.size)
        s.seek(s.getPos() + (size.value - ush.size));

    return true;
}

static void
loadUnitScoreRecord(Stream& ss,
                    int max_id,
                    GUnitScores& get(GUniverse&, int),
                    GUnitScoreDefinitions& defs,
                    GUniverse& univ)
{
    TUnitScoreHeader ush;
    if (readUnitScoreHeader(ss, ush)) {
        /* Read content */
        char data[6];
        GUnitScoreDefinitions::index_t index = defs.addScoreDefinition(ush);
        while (ss.read(data, 6) == 6) {
            int16_t id    = getInt16(data);
            int16_t score = getInt16(data+2);
            int16_t turn  = getInt16(data+4);
            if (id > 0 && id <= max_id)
                get(univ, id).mergeScore(index, score, turn);
        }
    } else {
        console.write(LOG_WARN, _("Unit score record is invalid"));
    }
}

/** Load starchart database.
    \param s   Stream to read from
    \param trn Turn to populate with information
    \param acceptProperties true to accept properties, false to skip them.
               Properties are not stored in the trn proper, so this offers
               to ignore properties from the file, avoiding overwriting
               global properties. */
void
loadChartDatabase(Stream& s, GGameTurn& trn, bool acceptProperties)
{
    using namespace std;

    int ignored_entries = 0;

    // Read header
    char header[16];
    s.seek(0);
    s.readT(header, sizeof(header));
    if (memcmp(header, "CCchart\032", 8) != 0)
        throw FileFormatException(s, _("File is missing required signature"));

    int16_t turn              = getInt16(header, 8);
    uint16_t start_offset     = getUint16(header, 10);
    uint16_t num_planet_props = getUint16(header, 12);
    uint16_t num_ship_props   = getUint16(header, 14);

    // Read property names
    IntVariableNames db_planet_props, db_ship_props;
    db_planet_props.load(s, num_planet_props);
    db_ship_props.load(s, num_ship_props);
    trn.addDatabaseTurn(turn);

    s.seek(start_offset);

    // Main loop
    GDrawingAtomMap atom_translation;
    GUniverse& univ = trn.getCurrentUniverse();

    char chunk_hdr[6];
    while (s.read(chunk_hdr, sizeof(chunk_hdr)) == sizeof(chunk_hdr)) {
        const int16_t type = getInt16(chunk_hdr, 0);
        int32_t size = getInt32(chunk_hdr, 2);        /* not const to allow count-down loops */
        const int32_t start_pos = s.getPos();
        const int32_t end_pos = start_pos + size;

        switch (type) {
         case rPlanetHistory: {
            /* a single planet record. Can have variable size, as we
               store more fields than PCC 1.x, and possibly extend it
               in the future. However, we expect a minimum size. */
            if (size >= 93) {
                /* 93 = size of planet data, plus timestamps */
                char tmp[TDbPlanet::size];
                zeroFill(tmp);
                if (size > TDbPlanet::size)
                    size = TDbPlanet::size;
                s.readT(tmp, size);

                TDbPlanet dbp;
                dbp.known_to_have_natives = 0;
                getPartialStructure(tmp, dbp, size);
                if (univ.isValidPlanetId(dbp.planet.planet_id))
                    univ.getPlanet(dbp.planet.planet_id).addHistoryData(dbp);
            } else {
                /* no known program writes these files */
            }
            break;
         }
         case rShipHistory: {
            if (size >= TDbShip::size) {
                /* Single ship history entry */
                TDbShip ship;
                getStructureT(s, ship);
                if (univ.isValidShipId(ship.ship.ship_id))
                    univ.getShip(ship.ship.ship_id).addShipHistoryData(ship);
            }
            break;
         }
         case rShipTrack: {
            if (size >= TDbShipTrack::size) {
                /* One TDbShipTrack, plus many TDbShipTrackEntry's */
                TDbShipTrack header;
                TDbShipTrackEntry entry;
                getStructureT(s, header);
                size -= header.size;
                if (univ.isValidShipId(header.id)) {
                    int turn = header.ref_turn;
                    while (size >= TDbShipTrackEntry::size) {
                        getStructureT(s, entry);
                        size -= entry.size;
                        univ.getShip(header.id).addShipTrackEntry(entry, turn);
                        --turn;
                    }
                }
            }
            break;
         }
         case rMinefield:
            while (size >= TDbMinefield::size) {
                /* read it */
                TDbMinefield dbm;
                getStructureT(s, dbm);
                size -= TDbMinefield::size;

                /* enter into database */
                TMinefieldReport report;
                report.x               = dbm.x;
                report.y               = dbm.y;
                report.owner           = dbm.owner;
                report.web             = dbm.type;
                report.units_known     = true;
                report.type_known      = true;
                report.radius_or_units = dbm.units;
                report.turn            = dbm.turn;
                report.why             = GMinefield::mfa_None;
                univ.ty_minefields.addMinefieldReport(dbm.id, report);
            }
            break;
         case rPainting: {
            SubStream ss(s);
            ss.setRange(start_pos, size);
            univ.getDrawings().load(ss, atom_translation, 0);  /* 0 means no expiration! */
            atom_translation.clear();
            break;
         }
         case rAutoBuild: {
            int id = 0;
            while (size >= TAutobuildSettings::size && id < NUM_PLANETS) {
                /* read it */
                TAutobuildSettings abs;
                getStructureT(s, abs);
                size -= TAutobuildSettings::size;

                /* enter into database */
                ++id;
                univ.getPlanet(id).setAutobuild(abs);
            }
            if (id != NUM_PLANETS || size != 0)
                console.write(LOG_WARN, _("Autobuild record has unexpected size"));
            break;
         }
         case rShipProperty:
            if (acceptProperties) {
                SubStream ss(s);
                ss.setRange(start_pos, size);
                loadPropertyRecord(ss, NUM_SHIPS, getShipProperties, db_ship_props, ship_property_names);
            }
            break;

         case rPlanetProperty:
            if (acceptProperties) {
                SubStream ss(s);
                ss.setRange(start_pos, size);
                loadPropertyRecord(ss, NUM_PLANETS, getPlanetProperties, db_planet_props, planet_property_names);
            }
            break;

         case rShipScore: {
            SubStream ss(s);
            ss.setRange(start_pos, size);
            loadUnitScoreRecord(ss, NUM_SHIPS, getShipScores, ship_score_definitions, univ);
            break;
         }
         case rPlanetScore: {
            SubStream ss(s);
            ss.setRange(start_pos, size);
            loadUnitScoreRecord(ss, NUM_PLANETS, getPlanetScores, planet_score_definitions, univ);
            break;
         }
         case rPaintingTags: {
            if (!atom_translation.isEmpty())
                console.write(LOG_WARN, _("Text record appears at unexpected place"));
            SubStream ss(s);
            ss.setRange(start_pos, size);
            atom_translation.clear();
            atom_translation.load(ss);
            break;
         }
         case rUfoHistory: {
            if (size >= TDbUfo::size) {
                TDbUfo ufo;
                getStructureT(s, ufo);
                univ.ty_ufos.addHistoryData(ufo);
            }
            break;
         }
         default:
            ++ignored_entries;
        }

        s.seek(end_pos);
    }
    if (ignored_entries != 0)
        console.write(LOG_INFO, format(_("%d database record%!1{s have%| has%} been ignored"), ignored_entries));
}

/*! \name Saving Starchart Database */
//@{
/// Status of saving.
typedef uint32_t RecordState;

/// Start a new record.
static void
startRecord(Stream& s, int type, RecordState& state)
{
    state = s.getPos();

    char header[6];
    storeInt16(header, type);
    storeInt32(header+2, 0);
    s.writeT(header, sizeof(header));
}

/// End a record.
static void
endRecord(Stream& s, RecordState& state)
{
    // state points to beginning of header. Since then, we wrote 6+N bytes.
    uint32_t pos = s.getPos();
    uint32_t bytes_in_chunk = s.getPos() - state - 6;

    // write it
    char tmp[4];
    storeInt32(tmp, bytes_in_chunk);
    s.seek(state + 2);
    s.writeT(tmp, sizeof(tmp));

    // go back
    s.seek(pos);
}

static void
writePropertyRecord(Stream& s, int type, int id, IntDataSegment& data)
{
    uint32_t n = data.getNumberOfUsedSlots();
    if (n != 0) {
        char header[4];
        storeInt16(header, id);
        storeInt16(header+2, n); // FIXME: overflow?

        RecordState rs;
        startRecord(s, type, rs);
        s.writeT(header, 4);
        data.save(s, n);
        endRecord(s, rs);
    }
}

/** Write unit score record.
    \param s Stream to write to
    \param type Type of chart.cc record to create
    \param max_id Object Ids are in range [1,max_id]
    \param defs Unit score definitions
    \param get Function to access an object's scores
    \param univ Universe to work on */
static void
writeUnitScores(Stream& s,
                int type,
                int max_id,
                const GUnitScoreDefinitions& defs,
                GUnitScores& get(GUniverse&, int),
                const GUniverse& univ)
{
    // FIXME: it would make sense to drop empty records. This implementation will
    // write them out, keeping outdated score definitions around for ages.
    char subheader[2];
    storeInt16(subheader, sizeof(TUnitScoreHeader));
    for (GUnitScoreDefinitions::index_t index = 0; index < defs.getNumScores(); ++index) {
        RecordState rs;
        startRecord(s, type, rs);
        s.writeT(subheader, 2);
        storeStructureT(s, defs.getScoreDefinition(index));
        for (int oid = 1; oid <= max_id; ++oid) {
            const GUnitScores& p = get(const_cast<GUniverse&>(univ), oid);
            const int16_t turn = p.getTurn(index);
            if (turn > 0) {
                char data[6];
                storeInt16(data, oid);
                storeInt16(data+2, p.getScore(index));
                storeInt16(data+4, turn);
                s.writeT(data, 6);
            }
        }
        endRecord(s, rs);
    }
}

/** Save starchart database.
    \param s   Stream to save into
    \param trn Turn to save */
void
saveChartDatabase(Stream& s, GGameTurn& trn)
{
    // Write preliminary header
    char header[16] = "ZZchart\032";
    storeInt16(header+8, trn.getTurnNumber());
    storeInt16(header+10, 0 /* start offset */);
    storeInt16(header+12, planet_property_names.getNumNames());
    storeInt16(header+14, ship_property_names.getNumNames());
    s.writeT(header, sizeof(header));
    planet_property_names.save(s, planet_property_names.getNumNames());
    ship_property_names.save(s, ship_property_names.getNumNames());

    // Prepare final header
    header[0] = header[1] = 'C';   // fix magic number
    storeInt16(header+10, s.getPos());

    const GUniverse& univ = trn.getCurrentUniverse();
    RecordState rs;

    /* Write the drawings. For drawings that have atoms attached, we
       must store the atom mapping first, to be able to restore it. 
       Whereas PCC 1.x stores the actual atom values verbatim, we
       emulate that to support our larger value range. This logic is
       entirely hidden in computeAtomMap(). */
    GDrawingAtomMap map;
    const GDrawingContainer& drawings = univ.getDrawings();
    drawings.computeAtomMap(map, GDrawingContainer::IncludeInvisible);

    if (!map.isEmpty()) {
        startRecord(s, rPaintingTags, rs);
        map.save(s);
        endRecord(s, rs);
    }
    startRecord(s, rPainting, rs);
    drawings.save(s, map);
    endRecord(s, rs);

    /* Write the minefields. It seems PCC 1.x stores only one
       minefield per record. This record has been specified as being
       capable of holding many since ever, and all other CHART.CC
       programs seem to handle that just fine, so we'll happily store
       them all in one. */
    startRecord(s, rMinefield, rs);
    for (GMinefieldType::const_iterator i = univ.ty_minefields.begin(); i != univ.ty_minefields.end(); ++i) {
        const GMinefield& mf = *i;
        TDbMinefield dbm;
        dbm.id    = mf.getId();
        dbm.x     = mf.getPos().x;
        dbm.y     = mf.getPos().y;
        dbm.owner = mf.getOwner();
        dbm.units = mf.getUnitsLastSeen();
        dbm.type  = mf.isWeb();
        dbm.turn  = mf.getTurnLastSeen();
        storeStructureT(s, dbm);
    }
    endRecord(s, rs);

    /* Write autobuild settings */
    startRecord(s, rAutoBuild, rs);
    for (int i = 1; i <= NUM_PLANETS; ++i) {
        TAutobuildSettings abs;
        univ.getPlanet(i).getAutobuild(abs);
        storeStructureT(s, abs);
    }
    endRecord(s, rs);

    /* Write planets */
    for (int i = 1; i <= NUM_PLANETS; ++i) {
        const GPlanet& p = univ.getPlanet(i);
        if (p.hasAnyPlanetData()) {
            TDbPlanet dbp;
            univ.getPlanet(i).getHistoryData(dbp);
            startRecord(s, rPlanetHistory, rs);
            storeStructureT(s, dbp);
            endRecord(s, rs);
        }
        writePropertyRecord(s, rPlanetProperty, i, getPlanetProperties(i));
    }

    /* Write ships */
    for (int i = 1; i <= NUM_SHIPS; ++i) {
        const GShip& sh = univ.getShip(i);
        // History Data
        if (sh.hasAnyShipData()) {
            TDbShip dbs;
            sh.getShipHistoryData(dbs);
            startRecord(s, rShipHistory, rs);
            storeStructureT(s, dbs);
            endRecord(s, rs);
        }

        // Track data
        TDbShipTrack th;
        int n;
        if ((n = sh.getShipTrackHeader(th)) != 0) {
            startRecord(s, rShipTrack, rs);
            storeStructureT(s, th);
            for (int i = 0; i < n; ++i) {
                TDbShipTrackEntry te;
                sh.getShipTrackEntry(te, i);
                storeStructureT(s, te);
            }
            endRecord(s, rs);
        }

        // Property data
        writePropertyRecord(s, rShipProperty, i, getShipProperties(i));
    }

    /* Write unit scores */
    writeUnitScores(s, rPlanetScore, NUM_PLANETS, planet_score_definitions, getPlanetScores, univ);
    writeUnitScores(s, rShipScore,   NUM_SHIPS,   ship_score_definitions,   getShipScores,   univ);

    /* Write Ufos */
    for (GUfoType::const_iterator i = univ.ty_ufos.begin(); i != univ.ty_ufos.end(); ++i) {
        if (i->isStoredInHistory()) {
            TDbUfo ufo;
            startRecord(s, rUfoHistory, rs);
            i->getHistoryData(ufo);
            storeStructureT(s, ufo);
            endRecord(s, rs);
        }
    }

    /* Write final header */
    s.seek(0);
    s.writeT(header, sizeof(header));
}


/*
 *  Score Accessors
 *
 *  These are intended to be used as function pointers to access the unit scores
 *  given a GUniverse.
 */

/** Get ship scores.
    \param univ Universe to work on
    \param id Return this ship's scores */
GUnitScores&
getShipScores(GUniverse& univ, int id)
{
    ASSERT(univ.isValidShipId(id));
    return univ.getShip(id).unit_scores;
}

/** Get planet scores.
    \param univ Universe to work on
    \param id Return this planet's scores */
GUnitScores&
getPlanetScores(GUniverse& univ, int id)
{
    ASSERT(univ.isValidPlanetId(id));
    return univ.getPlanet(id).unit_scores;
}
