/**
  *  \file game/msg.cc
  *  \brief Subspace Message Handling
  *
  *  Implementation of the abstract GMessagebox class.
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cctype>
#include "game/msg.h"
#include "game/struct.h"
#include "util/nls.h"
#include "util/except.h"
#include "game/pref.h"
#include "game/player.h"
#include "cpluslib/assert.h"
#include "util/recode.h"
#include "util/misc.h"
#include "io/io.h"
#include "game/racename.h"

/****************************** GMessagebox ******************************/

/*! \class GMessagebox
    This class provides an abstract interface to message boxes; basically
    everything that's needed to implement a standard message reader.
    In addition, it maintains a current position. Messages are indexed
    0-based. */
GMessagebox::GMessagebox()
    : current(0)
{ }

GMessagebox::~GMessagebox()
{ }

/** Set current message. If the index is out of range, it is
    automatically adjusted. If the change results in a change of the
    currently-selected message, raises sig_sel_changed. */
void
GMessagebox::setCurrent(int aindex)
{
    int count = getCount();
    if (aindex >= count)
        aindex = count-1;
    if (aindex < 0)
        aindex = 0;
    if (aindex != current) {
        current = aindex;
        sig_sel_changed.raise();
    }
}

/******************************* Utilities *******************************/

/** Parse a char array into a message. This applies a few fixups to
    the message:
    - rot13 is decoded, of course;
    - Dominate fixup: drop everything from NUL to EOL;
    - Winplan fixup (if rewrap is on), see below.
    - Recoding character sets.

    The message is returned in C++ format (lines separated by '\n'),
    no matter what that means in binary.

    Winplan fixup: older Winplans send messages with CR+LF line
    terminators, not CR as usual. Host messes these up because it
    thinks the LFs are text characters, and inserts additional CRs.
    Therefore, we change the meaning of the control characters such
    that LF means a linefeed and CR is ignored (because it was most
    likely added by HOST). However, headers contain CRs that are real
    linefeeds. Note that this might return lines longer than 40
    characters; display fixes those up.

    \param array    data
    \param size     length
    \param rewrap   true iff data is from inbox (and user wants rewrap),
                    false if from outbox

    More or less a copy of SendMsg::PChar2Edit, ReadMsg::DecodeMessage. */
string_t
getMessageFromArray(const char* array, std::size_t size, bool rewrap)
{
    string_t result;
    enum { None, Before, Inside } rewrap_status;

    /* Winplan fixup needed? */
    rewrap_status = None;
    if (rewrap) {
        for (std::size_t i = 0; i < size; ++i) {
            if (array[i] == 10 + 13) {
                rewrap_status = Before;
                break;
            }
        }
    }

    /* Strip trailing blanks */
    while (size && (array[size-1] == 32+13 || array[size-1] == 13+13 || array[size-1] == 13+10))
        --size;

    /* Convert message. */
    bool skipping = false;
    for (std::size_t i = 0; i < size; ++i) {
        char c = array[i] - 13;
        switch (c) {
         case 13:
            if (rewrap_status != Inside) {
                if (rewrap_status == Before
                    && result.size() && result[result.size()-1] == '\n')
                {
                    // a blank line, i.e. end of the headers
                    rewrap_status = Inside;
                }
                result += '\n';
            }
            skipping = false;
            break;
         case 10:
            if (rewrap_status == Inside)
                result += '\n';
            skipping = false;
            break;
         case 0:
            skipping = true;
            break;
         default:
            if (!skipping)
                result += c;
        }
    }
    return convertGameToUtf8(result);
}
