/**
  *  \file game/msgparse.cc
  *  \brief Message Parser
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/msgparse.h"
#include "cpluslib/ptrvector.h"
#include "cpluslib/string.h"
#include "util/nls.h"
#include "util/console.h"
#include "io/textfile.h"
#include "cpluslib/format.h"
#include "util/misc.h"
#include "game/struct.h"
#include "game/racename.h"

/** Parse a Match instruction. Extracts an optional scope modifier to build the final opcode,
    and generates the instruction.
    - "+n,text": search n-th line below
    - "-n,text": search n-th line above
    - "=n,text": search line n
    \param tpl Message template
    \param opcode Base opcode
    \param line Line to process */
static void
parseCheckInstruction(GMessageTemplate* tpl, uint8_t opcode, string_t line)
{
    uint8_t scope = GMessageTemplate::sAny;
    int8_t offset = 0;
    int8_t scale  = 1;
    if (!line.empty()) {
        if (line[0] == '+') {
            scope = GMessageTemplate::sRelative;
        } else if (line[0] == '-') {
            scope = GMessageTemplate::sRelative;
            scale = -1;
        } else if (line[0] == '=') {
            scope = GMessageTemplate::sFixed;
        }
        if (scope != GMessageTemplate::sAny) {
            string_t::size_type i = 1;
            while (i < line.size() && line[i] >= '0' && line[i] <= '9') {
                offset = 10*offset + line[i] - '0';
                ++i;
            }
            if (i < line.size() && line[i] == ',') {
                /* It's valid, so finish it */
                offset *= scale;
                ++i;
                while (i < line.size() && line[i] == ' ')
                    ++i;
                line.erase(0, i);
            } else {
                /* Invalid, back out */
                scope = GMessageTemplate::sAny;
            }
        }
    }
    tpl->addCheckInstruction(opcode + scope, offset, line);
}

/** Check that the template so far is sensible and generate warnings.
    \param tpl Template
    \param tf  Text file we're reading from (used to obtain identification information for warnings) */
static void
checkTemplate(const GMessageTemplate* tpl, TextFile& tf, int startingLine)
{
    /* No problem if there is no current template */
    if (tpl == 0)
        return;

    /* Check number of variables */
    int nvar  = tpl->getNumVariables();
    int nwild = tpl->getNumWildcards();
    if (nvar != nwild) {
        console.write(LOG_ERROR, format(_("%s:%d: number of variables (%d) does not match number of produced values (%d)"),
                                        tf.getName(), startingLine, nvar, nwild));
    }

    /* Check number of restrictions */
    if (tpl->getNumRestrictions() == 0) {
        console.write(LOG_ERROR, format(_("%s:%d: template will match every message"), tf.getName(), startingLine));
    }
}

/***************************** GMessageParser ****************************/

/** Constructor. Makes a blank message parser. */
GMessageParser::GMessageParser()
{ }

GMessageParser::~GMessageParser()
{ }

/** Load message parser information from a file. Builds message templates and adds
    them to this parser.
    \param in file opened for reading */
void
GMessageParser::load(Stream& in)
{
    TextFile tf(in);
    string_t line;
    GMessageTemplate* current_template = 0;
    int current_template_line = 0;

    while (tf.getLine(line)) {
        line = strTrim(line);
        if (line.empty() || line[0] == ';')
            continue;

        string_t::size_type p = line.find_first_of("=,");
        if (p == string_t::npos) {
            console.write(LOG_ERROR, format(_("%s:%d: missing delimiter"), in.getName(), tf.getLineNr()));
        } else if (line[p] == ',') {
            /* Check old template */
            checkTemplate(current_template, tf, current_template_line);

            /* Start a new template */
            GMessageObject mo = MessageNoObject;
            string_t kind = strTrim(line.substr(0, p));
            if (stringMatch("Minefield", kind)) {
                mo = MessageMinefield;
            } else if (stringMatch("Planet", kind)) {
                mo = MessagePlanet;
            } else if (stringMatch("Base", kind)) {
                mo = MessageStarbase;
            } else if (stringMatch("PLAYerscore", kind)) {
                mo = MessagePlayerScore;
            } else if (stringMatch("Ship", kind)) {
                mo = MessageShip;
            } else if (stringMatch("Ionstorm", kind)) {
                mo = MessageIonStorm;
            } else if (stringMatch("Configuration", kind)) {
                mo = MessageConfig;
            } else if (stringMatch("Explosion", kind)) {
                mo = MessageExplosion;
            } else if (stringMatch("Alliance", kind)) {
                mo = MessageAlliance;
            }
            if (mo == MessageNoObject) {
                console.write(LOG_ERROR, format(_("%s:%d: unknown object kind"), in.getName(), tf.getLineNr()));
                current_template = 0;
            } else {
                current_template = new GMessageTemplate(mo, strTrim(line.substr(p+1)));
                current_template_line = tf.getLineNr();
                message_templates.push_back(current_template);
            }
        } else {
            /* Assignments */
            if (!current_template)
                continue;

            string_t lhs = strTrim(line.substr(0, p));
            line = strTrim(line.substr(p+1));
            if (stringMatch("KInd", lhs)) {
                if (line.size())
                    current_template->addMatchInstruction(GMessageTemplate::iMatchKind, line[0]);
            } else if (stringMatch("SUbid", lhs)) {
                if (line.size())
                    current_template->addMatchInstruction(GMessageTemplate::iMatchSubId, line[0]);
            } else if (stringMatch("BIgid", lhs)) {
                int n;
                if (stringValue(line, n)) {
                    current_template->addMatchInstruction(GMessageTemplate::iMatchBigId, n);
                } else {
                    console.write(LOG_ERROR, format(_("%s:%d: invalid number"), in.getName(), tf.getLineNr()));
                }
            } else if (stringMatch("CHeck", lhs)) {
                parseCheckInstruction(current_template, GMessageTemplate::iCheck, line);
            } else if (stringMatch("FAil", lhs)) {
                parseCheckInstruction(current_template, GMessageTemplate::iFail, line);
            } else if (stringMatch("FInd", lhs)) {
                parseCheckInstruction(current_template, GMessageTemplate::iFind, line);
            } else if (stringMatch("PArse", lhs)) {
                parseCheckInstruction(current_template, GMessageTemplate::iParse, line);
            } else if (stringMatch("ARray", lhs)) {
                parseCheckInstruction(current_template, GMessageTemplate::iArray, line);
            } else if (stringMatch("VAlues", lhs)) {
                current_template->addValueInstruction(GMessageTemplate::iValue, line);
            } else if (stringMatch("ASsign", lhs)) {
                current_template->addVariables(line);
            } else if (stringMatch("COntinue", lhs)) {
                current_template->setContinueFlag(stringMatch("Yes", line));
            } else {
                console.write(LOG_WARN, format(_("%s:%d: unknown keyword"), in.getName(), tf.getLineNr()));
            }
        }
    }
    checkTemplate(current_template, tf, current_template_line);
}

/** Discard all templates. */
void
GMessageParser::done()
{
    message_templates.clear();
}



#if 1
void
GMessageParser::dump()
{
    for (ptr_vector<GMessageTemplate>::iterator i = message_templates.begin(); i != message_templates.end(); ++i) {
        console << "---- Template '" << (*i)->getTemplateName() << "' ----\n"
                << "continue_flag = " << ((*i)->getContinueFlag() ? "on" : "off") << "\n\n";
        (*i)->dump();
    }
}
#endif


/** Convert pre-parsed yes/no array into an array of offers. */
static void
generateSimpleAllies(GPlayerArray<GAllianceOffer::OfferType>& out, string_t value)
{
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        string_t item = strTrim(strFirst(value, ","));
        if (stringMatch("Yes", item)) {
            out[i] = GAllianceOffer::Yes;
        } else if (stringMatch("No", item)) {
            out[i] = GAllianceOffer::No;
        } else if (stringMatch("Conditional", item)) {
            out[i] = GAllianceOffer::Conditional;
        } else {
            // ignore (in particular, this branch is taken when the field is empty)
        }
        strRemove(value, ",");
    }
}

/** Generate allies from an array of "Race+!" elements. */
static void
generateFlagAllies(GAllianceOffer& out, string_t value)
{
    do {
        string_t item = strFirst(value, ",");

        // Parse off the alliance markers
        bool excl = false;
        bool plus = false;
        size_t n = item.size();
        while (n > 0) {
            if (item[n-1] == '+') {
                plus = true;
                --n;
            } else if (item[n-1] == '!') {
                excl = true;
                --n;
            } else if (item[n-1] == ' ' || item[n-1] == ':') {
                --n;
            } else {
                break;
            }
        }

        // What remains is a race name, I hope.
        //   !     => this race has offered something to us
        //   +     => we have offered something
        item.erase(n);
        for (int i = 1; i <= NUM_OWNERS; ++i) {
            if (strCaseCmp(item, strTrim(host_racenames.getAdjName(i))) == 0) {
                out.theirOffer[i] = excl ? GAllianceOffer::Yes : GAllianceOffer::No;
                out.oldOffer[i]   = plus ? GAllianceOffer::Yes : GAllianceOffer::No;
                break;
            }
        }
    } while (strRemove(value, ","));
}


/** Generate output for one matching message template. This generates a new GMessageInformation
    record, or extends an existing one.
    \param values     [in] values produced by the template
    \param tpl        [in] template which matched
    \param turn       [in] message turn number
    \param info_out   [out] Information will be appended here */
static void
generateOutput(const std::vector<string_t>& values, const GMessageTemplate& tpl, const int turn, ptr_vector<GMessageInformation>& info_out)
{
    GMessageInformation* info;

    /* Figure out process limit. We cannot process values that have no variable,
       nor variables without values. */
    uint_t processLimit = values.size();
    if (processLimit > uint_t(tpl.getNumVariables()))
        processLimit = uint_t(tpl.getNumVariables());

    /* Figure out Id number. */
    int id = 0;
    int limit;
    bool mergeable = false;
    uint_t skipSlot = values.size()+1;
    switch (GMessageObject obj = tpl.getMessageObject()) {
     case MessageShip:
     case MessageMinefield:
     case MessagePlanet:
     case MessageStarbase:
     case MessageIonStorm:
        /* Those are identified by a mandatory Id */
        skipSlot = tpl.getVariableSlotByName("ID");
        if (skipSlot < values.size()) {
            id = tpl.getIntValue(values[skipSlot]);
        }
        if (!id) {
            /* Only complain about missing Id when we actually produced some data.
               Some templates produce just an Id, "just in case", to associate the
               message with an object or produce a marker. */
            if (values.size() > (skipSlot < values.size())) {
                console.write(LOG_ERROR, format(_("Message template \"%s\" did not produce Id number"), tpl.getTemplateName()));
            }
            return;
        }
        limit = (obj == MessageShip ? NUM_SHIPS
                 : obj == MessageMinefield ? 10000 /* FIXME: otherwise, there's no limit for minefields */
                 : obj == MessageIonStorm ? NUM_STORMS
                 : NUM_PLANETS);                   /* MessageStarbase, MessagePlanet */
        if (id <= 0 || id > limit) {
            console.write(LOG_ERROR, format(_("Message template \"%s\" produced out-of-range Id %d, ignoring"), tpl.getTemplateName(), id));
            return;
        }
        mergeable = true;
        break;

     case MessagePlayerScore:
        /* These can have an optional Id */
        skipSlot = tpl.getVariableSlotByName("ID");
        if (skipSlot < values.size()) {
            id = tpl.getIntValue(values[skipSlot]);
        }
        mergeable = (id != 0);
        break;

     case MessageConfig:
        /* Always mergeable */
        mergeable = true;
        break;

     case MessageAlliance:
        /* This one is special, see below */
        break;

     case MessageExplosion:
     case MessageNoObject:
        break;
    }

    /* Find out whether we can merge this item with the previous one. We can merge when
       the object kind permits merging, and they actually describe the same object. */
    if (mergeable
        && !info_out.empty()
        && info_out.back()->getObjectType() == tpl.getMessageObject()
        && info_out.back()->getObjectId() == id
        && info_out.back()->getTurnNumber() == turn)
    {
        /* Merge this */
        info = info_out.back();
    } else {
        /* Make new object */
        info = new GMessageInformation(tpl.getMessageObject(), id, turn);
        info_out.push_back(info);
    }

    /* Now produce the values */
    if (tpl.getMessageObject() == MessageAlliance) {
        /* Alliance case. Produce one GAllianceOffer and a name. */
        GAllianceOffer offer;
        string_t idStr;

        for (uint_t i = 0; i < processLimit; ++i) {
            const string_t varName = tpl.getVariableName(i);
            if (values[i].empty() || varName == "_" || varName.empty())
                continue;

            if (varName == "NAME") {
                idStr = values[i];
            } else if (varName == "FROM") {
                generateSimpleAllies(offer.theirOffer, values[i]);
            } else if (varName == "TO") {
                generateSimpleAllies(offer.oldOffer, values[i]);
            } else if (varName == "FLAGS") {
                generateFlagAllies(offer, values[i]);
            } else {
                console.write(LOG_ERROR, format(_("Message template \"%s\" generates unknown value \"%s\""),
                                                tpl.getTemplateName(), tpl.getVariableName(i)));
            }
        }

        if (idStr.empty()) {
            console.write(LOG_ERROR, format(_("Message template \"%s\" did not produce name, ignoring"), tpl.getTemplateName()));
            return;
        }
        info->addAllianceValue(idStr, offer);
    } else {
        /* Regular case */
        for (uint_t i = 0; i < processLimit; ++i) {
            /* Do not handle empty values. Those are generated, in particular,
               by array items. Also skip the Id field processed above. Finally,
               skip values named "_" (match placeholders). */
            const string_t varName = tpl.getVariableName(i);
            if (values[i].empty() || i == skipSlot || varName == "_" || varName.empty())
                continue;

            GMessageStringIndex msi;
            GMessageIntIndex mii;
            if (tpl.getMessageObject() == MessageConfig) {
                /* Configuration produces naked key/value pairs */
                info->addConfigValue(varName, values[i]);
            } else if (tpl.getMessageObject() == MessagePlayerScore && varName == "SCORE") {
                /* Score, this is an 11-element array */
                string_t all = values[i];
                int pl = 1;
                string_t::size_type n;
                while ((n = all.find(',')) != string_t::npos) {
                    if (n != 0)
                        info->addScoreValue(pl, tpl.getIntValue(all.substr(0, n)));
                    all.erase(0, n+1);
                    ++pl;
                }
                if (!all.empty())
                    info->addScoreValue(pl, tpl.getIntValue(all));
            } else if ((msi = getStringIndexFromKeyword(varName)) != ms_Max) {
                /* String value */
                info->addValue(msi, values[i]);
            } else if ((mii = getIntIndexFromKeyword(varName)) != mi_Max) {
                /* Integer value */
                info->addValue(mii, tpl.getIntValue(values[i]));
            } else if ((varName[0] == '+' || varName[0] == '-') && (mii = getIntIndexFromKeyword(varName.substr(1))) != mi_Max) {
                /* Relative integer */
                bool ok = false;
                for (GMessageInformation::iterator it = info->begin(); it != info->end(); ++it) {
                    if (GMessageIntValue* iv = dynamic_cast<GMessageIntValue*>(*it)) {
                        if (iv->getIndex() == mii) {
                            int32_t delta = tpl.getIntValue(values[i]);
                            if (varName[0] == '-')
                                delta = -delta;
                            iv->setValue(iv->getValue() + delta);
                            ok = true;
                            break;
                        }
                    }
                }
                if (!ok) {
                    console.write(LOG_ERROR, format(_("Message template \"%s\" modifies value \"%s\" which does not exist"),
                                                    tpl.getTemplateName(), tpl.getVariableName(i).substr(1)));
                }
            } else {
                /* What? */
                console.write(LOG_ERROR, format(_("Message template \"%s\" generates unknown value \"%s\""),
                                                tpl.getTemplateName(), tpl.getVariableName(i)));
            }
        }
    }
}

/** Parse a message, main entry point.
    \param theMessage [in] Message text
    \param player     [in] Player who got the message
    \param turn       [in] Current turn number
    \param info_out   [out] Information will be appended here */
void
GMessageParser::parseMessage(const string_t theMessage, const int player, const int turn, ptr_vector<GMessageInformation>& info_out)
{
    /* Step 1: split message into lines */
    GMessageLines lines;
    splitMessage(lines, theMessage);

    /* Step 2: parse all templates and gather information */
    for (ptr_vector<GMessageTemplate>::iterator i = message_templates.begin(); i != message_templates.end(); ++i) {
        std::vector<string_t> values;
        if ((*i)->match(lines, player, values)) {
            /* Matches. Produce output. */
            generateOutput(values, **i, turn - getMessageHeaderInformation(lines, MsgHdrAge), info_out);
            if (!(*i)->getContinueFlag())
                break;
        }
    }
}
