/**
  *  \file game/vcrobj.h
  *  \brief VCR Object
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_VCROBJ_H
#define PCC_V2_GAME_VCROBJ_H

#include "cpluslib/types.h"
#include "cpluslib/string.h"

class TRawVcr;

/** Object taking part in a VCR.
    This describes an object taking part in a fight, containing the basic object specs.
    It has a virtual destructor so that combat algorithms using more than these specs
    can derive from this class, add their own properties, and be still recognizable
    using RTTI. Methods are inline, not virtual, because this is also used at the core
    of VCR algorithms which are intended to be fast. */
class GVcrObject {
 public:
    GVcrObject();
    GVcrObject(const TRawVcr& raw, int side);
    virtual ~GVcrObject();

    void unpack(const TRawVcr& raw, int side);
    bool pack(TRawVcr& raw, int side);

    int getMass() const;
    int getShield() const;
    int getDamage() const;
    int getCrew() const;
    int getId() const;
    int getOwner() const;
    int getRace() const;
    int getPicture() const;
    int getHull() const;
    int getBeamType() const;
    int getBeamCount() const;
    int getTorpType() const;
    int getTorpCount() const;
    int getTorpLauncherCount() const;
    int getBayCount() const;
    int getFighterCount() const;
    int getExperienceLevel() const;
    int getBeamKillRate() const;
    int getBeamChargeRate() const;
    int getTorpMissRate() const;
    int getTorpChargeRate() const;
    int getCrewDefenseRate() const;

    bool isPlanet() const;
    const string_t& getName() const;

    void setMass(int mass);
    void setShield(int shield);
    void setDamage(int damage);
    void setCrew(int crew);
    void setId(int id);
    void setOwner(int owner);
    void setRace(int race_or_zero);
    void setPicture(int picture);
    void setHull(int hull_or_zero);
    void setBeamType(int beam_type);
    void setBeamCount(int beam_count);
    void setTorpType(int torp_type);
    void setTorpCount(int torp_count);
    void setTorpLauncherCount(int torp_lcount);
    void setBayCount(int bay_count);
    void setFighterCount(int fighter_count);
    void setExperienceLevel(int experience_level);
    void setBeamKillRate(int beam_kill_rate);
    void setBeamChargeRate(int beam_charge_rate);
    void setTorpMissRate(int torp_miss_rate);
    void setTorpChargeRate(int torp_charge_rate);
    void setCrewDefenseRate(int crew_defense_rate);
    void setIsPlanet(bool is_planet);
    void setName(const string_t& name);

    void addFighters(int n);
    void addTorpedoes(int n);
    void addBays(int n);
    void addMass(int n);

    void setGuessedHull();
    bool canBeHull(int hull_id) const;
    int  getGuessedHull() const;
    int  getGuessedShipPicture() const;

 private:
    typedef int value_t;
    /* Placing all POD members in a structure causes gcc-3.3 to generate better code
       for construction/assignment. */
    struct {
        value_t  mass;
        value_t  shield;
        value_t  damage;
        value_t  crew;
        value_t  id;
        value_t  owner;
        value_t  race_or_zero;
        value_t  picture;
        value_t  hull_or_zero;
        value_t  beam_type;
        value_t  beam_count;
        value_t  torp_type;
        value_t  torp_count;
        value_t  torp_lcount;
        value_t  bay_count;
        value_t  fighter_count;
        value_t  experience_level;
        bool     is_planet;

        // Nu extensions:
        value_t  beam_kill_rate;
        value_t  beam_charge_rate;
        value_t  torp_miss_rate;
        value_t  torp_charge_rate;
        value_t  crew_defense_rate;
    } pod;
    string_t name;
};



/******************************** Inlines ********************************/

inline int
GVcrObject::getMass() const
{
    return pod.mass;
}

inline int
GVcrObject::getShield() const
{
    return pod.shield;
}

inline int
GVcrObject::getDamage() const
{
    return pod.damage;
}

inline int
GVcrObject::getCrew() const
{
    return pod.crew;
}

inline int
GVcrObject::getId() const
{
    return pod.id;
}

inline int
GVcrObject::getOwner() const
{
    return pod.owner;
}

inline int
GVcrObject::getRace() const
{
    return pod.race_or_zero;
}

inline int
GVcrObject::getPicture() const
{
    return pod.picture;
}

inline int
GVcrObject::getHull() const
{
    return pod.hull_or_zero;
}

inline int
GVcrObject::getBeamType() const
{
    return pod.beam_type;
}

inline int
GVcrObject::getBeamCount() const
{
    return pod.beam_count;
}

inline int
GVcrObject::getTorpType() const
{
    return pod.torp_type;
}

inline int
GVcrObject::getTorpCount() const
{
    return pod.torp_count;
}

inline int
GVcrObject::getTorpLauncherCount() const
{
    return pod.torp_lcount;
}

inline int
GVcrObject::getBayCount() const
{
    return pod.bay_count;
}

inline int
GVcrObject::getFighterCount() const
{
    return pod.fighter_count;
}

inline int
GVcrObject::getExperienceLevel() const
{
    return pod.experience_level;
}

inline int
GVcrObject::getBeamKillRate() const
{
    return pod.beam_kill_rate;
}

inline int
GVcrObject::getBeamChargeRate() const
{
    return pod.beam_charge_rate;
}

inline int
GVcrObject::getTorpMissRate() const
{
    return pod.torp_miss_rate;
}

inline int
GVcrObject::getTorpChargeRate() const
{
    return pod.torp_charge_rate;
}

inline int
GVcrObject::getCrewDefenseRate() const
{
    return pod.crew_defense_rate;
}

inline bool
GVcrObject::isPlanet() const
{
    return pod.is_planet;
}

inline const string_t&
GVcrObject::getName() const
{
    return name;
}

inline void
GVcrObject::setMass(int mass)
{
    this->pod.mass = mass;
}

inline void
GVcrObject::setShield(int shield)
{
    this->pod.shield = shield;
}

inline void
GVcrObject::setDamage(int damage)
{
    this->pod.damage = damage;
}

inline void
GVcrObject::setCrew(int crew)
{
    this->pod.crew = crew;
}

inline void
GVcrObject::setId(int id)
{
    this->pod.id = id;
}

inline void
GVcrObject::setOwner(int owner)
{
    this->pod.owner = owner;
}

inline void
GVcrObject::setRace(int race_or_zero)
{
    this->pod.race_or_zero = race_or_zero;
}

inline void
GVcrObject::setPicture(int picture)
{
    this->pod.picture = picture;
}

inline void
GVcrObject::setHull(int hull_or_zero)
{
    this->pod.hull_or_zero = hull_or_zero;
}

inline void
GVcrObject::setBeamType(int beam_type)
{
    this->pod.beam_type = beam_type;
}

inline void
GVcrObject::setBeamCount(int beam_count)
{
    this->pod.beam_count = beam_count;
}

inline void
GVcrObject::setTorpType(int torp_type)
{
    this->pod.torp_type = torp_type;
}

inline void
GVcrObject::setTorpCount(int torp_count)
{
    this->pod.torp_count = torp_count;
}

inline void
GVcrObject::setTorpLauncherCount(int torp_lcount)
{
    this->pod.torp_lcount = torp_lcount;
}

inline void
GVcrObject::setBayCount(int bay_count)
{
    this->pod.bay_count = bay_count;
}

inline void
GVcrObject::setFighterCount(int fighter_count)
{
    this->pod.fighter_count = fighter_count;
}

inline void
GVcrObject::setExperienceLevel(int experience_level)
{
    this->pod.experience_level = experience_level;
}

inline void
GVcrObject::setBeamKillRate(int beam_kill_rate)
{
    this->pod.beam_kill_rate = beam_kill_rate;
}

inline void
GVcrObject::setBeamChargeRate(int beam_charge_rate)
{
    this->pod.beam_charge_rate = beam_charge_rate;
}

inline void
GVcrObject::setTorpMissRate(int torp_charge_rate)
{
    this->pod.torp_miss_rate = torp_charge_rate;
}

inline void
GVcrObject::setTorpChargeRate(int torp_charge_rate)
{
    this->pod.torp_charge_rate = torp_charge_rate;
}

inline void
GVcrObject::setCrewDefenseRate(int crew_defense_rate)
{
    this->pod.crew_defense_rate = crew_defense_rate;
}

inline void
GVcrObject::setIsPlanet(bool is_planet)
{
    this->pod.is_planet = is_planet;
}

inline void
GVcrObject::setName(const string_t& name)
{
    this->name = name;
}

inline void
GVcrObject::addFighters(int n)
{
    this->pod.fighter_count += n;
}

inline void
GVcrObject::addTorpedoes(int n)
{
    this->pod.torp_count += n;
}

inline void
GVcrObject::addBays(int n)
{
    this->pod.bay_count += n;
}

inline void
GVcrObject::addMass(int n)
{
    this->pod.mass += n;
}

#endif
