/**
  *  \file util/cmdline.cc
  *  \brief Command Line Handling
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include <ostream>
#include "util/cmdline.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "cpluslib/utf8.h"

/*! \class CommandLineParser
    Encapsulates the logic for parsing a command line into
    options and non-options. Derived classes implement the
    actual parameter interpretation.

    Our syntax:
    - options start with `-' or `--', e.g. `-display', `--help'.
      Single-dash and double-dash is identical.
    - an argument consisting of just two dashes stops interpretation
      of options and causes the rest to be interpreted as non-options.
    - everything else is a non-option. Note that `-' is a non-option,
      too; this is useful because it is usually used in place of a
      file name to mean stdin/stdout.

    With `traditional', it uses the traditional syntax, where
    a single dash introduces a block of combined options
    (`-blah' == `-b -l -a -h'). */

const char* CommandLineParser::prog;

/** Initialize.
    \param trad Enable traditional behaviour for short options */
CommandLineParser::CommandLineParser(bool trad)
    : argv(0), traditional(trad)
{ }

CommandLineParser::~CommandLineParser()
{ }

/** Internal: get next element from parameter vector. With
    "saturation" at end. */
const char*
CommandLineParser::getNext()
{
    if (*argv)
        return *argv++;
    else
        return 0;
}

/** Get argument. handleOption() can call this to request the current
    argument. */
const char*
CommandLineParser::getArg()
{
    if (!arg)
        arg = getNext();
    return arg;
}

/** Parse. This one does the actual work. */
void
CommandLineParser::parse(const char*const* aargv)
{
    argv = aargv;
    prog = getNext();
    parseStarted();
    while (const char* p = getNext()) {
        arg = 0;
        if (p[0] != '-' || p[1] == 0) {
            /* non-option: doesn't start with `-' or only one character */
            handleNonoption(p);
        } else if(p[1] == '-' && p[2] == 0) {
            /* option-stop `--': rest is non-options */
            while (const char* q = getNext())
                handleNonoption(q);
            break;
        } else {
            /* option. Skip at most two dashes */
            ++p;
            if (traditional && *p != '-') {
                /* traditional handling. `-blah' == `-b -l -a -h' */
                // FIXME: "-blah" == "-b lah" if -b takes an argument
                Utf8Reader optset(p, std::strlen(p));
                while (optset.hasMore()) {
                    string_t one;
                    appendUtf8(one, optset.get());
                    handleOption(one, false);
                }
            } else {
                /* X11/GNU style parsing. `-display' == `--display' */
                if (*p == '-')
                    ++p;
                const char* eq = std::strchr(p, '=');
                if (!eq) {
                    handleOption(p, false);
                } else {
                    arg = eq+1;
                    handleOption(string_t(p, eq-p), true);
                }
            }
        }
    }
    parseFinished();
}

/** Hook: called before parsing the command line.
    \default does nothing */
void
CommandLineParser::parseStarted()
{ }

/** Hook: called after successfully parsing the command line.
    \default does nothing */
void
CommandLineParser::parseFinished()
{ }

/** \fn CommandLineParser::handleOption(const string_t&, bool)
    Handle an option. The \c have_arg parameter is true when the option
    was specified with a parameter (`--param=foo'); this is so that
    the parser can reject options that ought not to take parameters
    but the user gave some. No matter how the argument was specified,
    handleOption() can request it using getArg(). */

/** \fn CommandLineParser::handleNonoption(const string_t&)
    Handle a non-option. */

/*********************** StandardCommandLineParser ***********************/

/*! \class StandardCommandLineParser
    \brief Standardized Command Line Parser

    This command line parser takes its orders from an array of Option
    structures. It also allows users to abbreviate the option names if
    they are unique, as in "-h" for "-help". */

/** Initialize.
    \param trad Enable traditional behaviour for short options */
StandardCommandLineParser::StandardCommandLineParser(bool trad)
    : CommandLineParser(trad)
{ }

StandardCommandLineParser::~StandardCommandLineParser()
{ }

/** Add set of options.
    \param opt   [in] Option definitions
    \param count [in] Number of options to add */
void
StandardCommandLineParser::addOptionSet(const Option* opt,
                                        std::size_t count)
{
    if (count)
        options.insert(options.end(), opt, opt + count);
}

void
StandardCommandLineParser::handleOption(const string_t& opt, bool have_arg)
{
    /* Find option */
    int index = -1;
    string_t error;
    for (std::vector<Option>::size_type i = 0; i < options.size(); ++i) {
        std::size_t len = std::strlen(options[i].name);
        if (opt.length() <= len && opt.compare(0, opt.length(), options[i].name, opt.length()) == 0) {
            /* Option matches. */
            if (index < 0) {
                index = i;
            } else {
                /* ambiguous */
                if (! error.length())
                    /* translator: this should yield a partial
                       sentence, we'll append a number of tokens of
                       the form ", `name'" to this text */
                    error = format(_("Option `%s' is ambiguous, could be `%s'"),
                                     opt, options[index].name);
                error += format(", `%s'", options[i].name);
            }
        }
    }
    if (error.length()) {
        handleError(error);
    } else if (index < 0) {
        handleError(format(_("Unknown option `%s'"), opt));
    } else if (have_arg && !options[index].takes_arg) {
        handleError(format(_("Option `%s' does not take an argument"), opt));
    } else if (options[index].takes_arg && !getArg()) {
        handleError(format(_("Option `%s' requires an argument"), opt));
    } else {
        options[index].handler(*this, options[index]);
    }
}

/** Print a standard help message. */
void
StandardCommandLineParser::printHelp(std::ostream& os)
{
    for (std::vector<Option>::iterator i = options.begin(); i != options.end(); ++i)
        if (i->help)
            os << _(i->help) << std::endl;
}

/** Handle error.
    \default throws CommandLineException */
void
StandardCommandLineParser::handleError(const string_t& msg)
{
    throw CommandLineException(msg);
}
